use crate::array::growable::{Growable, GrowableArray};
use crate::array::ops::full::FullNull;
use crate::array::DataArray;
use crate::datatypes::logical::LogicalArrayImpl;
use crate::datatypes::{BooleanArray, DaftLogicalType, DaftPhysicalType};
use crate::DataType;
use arrow2::array::Array;
use common_error::DaftResult;

use super::as_arrow::AsArrow;

fn generic_if_else<'a, T: GrowableArray<'a> + FullNull + Clone>(
    predicate: &BooleanArray,
    name: &str,
    lhs: &'a T,
    rhs: &'a T,
    dtype: &DataType,
    lhs_len: usize,
    rhs_len: usize,
) -> DaftResult<T> {
    // CASE 1: Broadcast predicate
    if predicate.len() == 1 {
        return match predicate.get(0) {
            None => Ok(T::full_null(name, dtype, lhs_len)),
            Some(predicate_scalar_value) => {
                if predicate_scalar_value {
                    Ok(lhs.clone())
                } else {
                    Ok(rhs.clone())
                }
            }
        };
    }

    // Build the result using a Growable
    let predicate = predicate.as_arrow();
    let mut growable = T::make_growable(
        name.to_string(),
        dtype,
        vec![lhs, rhs],
        predicate.len(),
        predicate.null_count() > 0, // If predicate has nulls, we will need to append nulls to growable
    );
    // CASE 2: predicate is not broadcastable, and contains nulls
    //
    // NOTE: we iterate through the predicate+validity which is slightly slower than using a SlicesIterator approach.
    // An alternative is to naively apply the raw predicate values (without considering validity), and then applying
    // validity afterwards. However this could use more memory than required, since we copy even values that will
    // later then be considered null.
    if predicate.null_count() > 0 {
        for (i, pred) in predicate.into_iter().enumerate() {
            match pred {
                None => {
                    growable.add_nulls(1);
                }
                Some(true) => {
                    let idx = if lhs_len == 1 { 0 } else { i };
                    growable.extend(0, idx, 1);
                }
                Some(false) => {
                    let idx = if rhs_len == 1 { 0 } else { i };
                    growable.extend(1, idx, 1);
                }
            }
        }
        growable.build()

    // CASE 3: predicate is not broadcastable, and does not contain nulls
    } else {
        // Helper to extend the growable, taking into account broadcast semantics
        let (broadcast_lhs, broadcast_rhs) = (lhs_len == 1, rhs_len == 1);
        let mut extend = |pred: bool, start, len| {
            match (pred, broadcast_lhs, broadcast_rhs) {
                (true, false, _) => {
                    growable.extend(0, start, len);
                }
                (false, _, false) => {
                    growable.extend(1, start, len);
                }
                (true, true, _) => {
                    for _ in 0..len {
                        growable.extend(0, 0, 1);
                    }
                }
                (false, _, true) => {
                    for _ in 0..len {
                        growable.extend(1, 0, 1);
                    }
                }
            };
        };

        // Iterate through the predicate using SlicesIterator, which is a much faster way of iterating through a bitmap
        let mut start_falsy = 0;
        let mut total_len = 0;
        for (start, len) in arrow2::bitmap::utils::SlicesIterator::new(predicate.values()) {
            if start != start_falsy {
                extend(false, start_falsy, start - start_falsy);
                total_len += start - start_falsy;
            };
            extend(true, start, len);
            total_len += len;
            start_falsy = start + len;
        }
        if total_len != predicate.len() {
            extend(false, total_len, predicate.len() - total_len);
        }
        growable.build()
    }
}

impl<'a, T> DataArray<T>
where
    T: DaftPhysicalType + 'static,
    DataArray<T>: GrowableArray<'a>,
{
    pub fn if_else(
        &'a self,
        other: &'a DataArray<T>,
        predicate: &BooleanArray,
    ) -> DaftResult<DataArray<T>> {
        generic_if_else(
            predicate,
            self.name(),
            self,
            other,
            self.data_type(),
            self.len(),
            other.len(),
        )
    }
}

impl<'a, L> LogicalArrayImpl<L, DataArray<L::PhysicalType>>
where
    L: DaftLogicalType,
    LogicalArrayImpl<L, DataArray<L::PhysicalType>>: GrowableArray<'a>,
    LogicalArrayImpl<L, DataArray<L::PhysicalType>>: FullNull,
{
    pub fn if_else(
        &'a self,
        other: &'a LogicalArrayImpl<L, DataArray<L::PhysicalType>>,
        predicate: &BooleanArray,
    ) -> DaftResult<LogicalArrayImpl<L, DataArray<L::PhysicalType>>> {
        generic_if_else(
            predicate,
            self.name(),
            self,
            other,
            self.data_type(),
            self.len(),
            other.len(),
        )
    }
}
